@@ -25,6 +25,9 @@ import numpy as np
 
 import gridded.pyugrid.read_netcdf as read_netcdf
 from gridded.pyugrid.util import point_in_tri
+
+from gridded.utilities import get_writable_dataset
+
 #from gridded.pyugrid.uvar import UVar
 
 # __all__ = ['UGrid',
@@ -113,6 +116,11 @@ class UGrid(object):
         mesh section in the convention for further details.
         """
 
+        if ((nodes is not None) and
+            ((node_lon is not None) or
+             (node_lat is not None))):
+            raise TypeError("You need to provide a single nodes array "
+                            "or node_lon and node_lat")
         if nodes is None:
             if node_lon is not None and node_lat is not None:
                 nodes = np.ma.column_stack((node_lon, node_lat))
@@ -140,7 +148,9 @@ class UGrid(object):
         # A kdtree is used to locate nodes.
         # It will be created if/when it is needed.
         self._kdtree = None
-        self._tree = None
+        self._cell_tree = None
+        self._ind_memo_dict = OrderedDict()
+        self._alpha_memo_dict = OrderedDict()
 
     @classmethod
     def from_ncfile(klass, nc_url, mesh_name=None):  # , load_data=False):
@@ -388,24 +398,25 @@ class UGrid(object):
 
         None otherwise.
         """
-        try:
-            loc = data.location
-            if loc == "face":
-                # FIXME: should we check the array size in this case?
-                return "faces"
-        except AttributeError:
-            pass # try checking array size
-        # fixme: should use UGRID compliant nc_attributes if possible
+        # We should never be caling infer_locations is it was already defined
+        # try:
+        #     loc = data.location
+        #     if loc == "face":
+        #         # FIXME: should we check the array size in this case?
+        #         return "face"
+        # except AttributeError:
+        #     pass # try checking array size
+        # # fixme: should use UGRID compliant nc_attributes if possible
         try:
             size = data.shape[-1]
         except IndexError:
             return None  # Variable has a size-zero data array
         if size == self.nodes.shape[0]:
-            return 'nodes'
+            return 'node'
         if self.faces is not None and size == self.faces.shape[0]:
-            return 'faces'
+            return 'face'
         if self.boundaries is not None and size == self.boundaries.shape[0]:
-            return 'boundaries'
+            return 'boundary'
         return None
 
     # def add_data(self, uvar):
@@ -497,7 +508,13 @@ class UGrid(object):
 
     def _build_kdtree(self):
         # Only import if it's used.
-        from scipy.spatial import cKDTree
+        try:
+            from scipy.spatial import cKDTree
+        except ImportError:
+            raise ImportError("The scipy package is required to use "
+                              "UGrid.locate_nodes\n"
+                              " -- nearest neighbor interpolation")
+
         self._kdtree = cKDTree(self.nodes)
 
     def _hash_of_pts(self, points):
@@ -560,8 +577,6 @@ class UGrid(object):
         points = np.asarray(points, dtype=np.float64)
         just_one = (points.ndim == 1)
         points.shape = (-1, 2)
-        if not hasattr(self, '_ind_memo_dict'):
-            self._ind_memo_dict = OrderedDict()
 
         if _memo:
             if _hash is None:
@@ -576,9 +591,9 @@ class UGrid(object):
             except ImportError:
                 raise ImportError("the cell_tree2d package must be installed to use the celltree search:\n"
                                   "https://github.com/NOAA-ORR-ERD/cell_tree2d/")
-            if self._tree is None:
+            if self._cell_tree is None:
                 self.build_celltree()
-            indices = self._tree.locate(points)
+            indices = self._cell_tree.locate(points)
         elif method == 'simple':
             indices = np.zeros((points.shape[0]), dtype=IND_DT)
             for n, point in enumerate(points):
@@ -609,7 +624,7 @@ class UGrid(object):
         if self.nodes is None or self.faces is None:
             raise ValueError(
                 "Nodes and faces must be defined in order to create and use CellTree")
-        self._tree = CellTree(self.nodes, self.faces)
+        self._cell_tree = CellTree(self.nodes, self.faces)
 
     def interpolation_alphas(self, points, indices=None, _copy=False, _memo=True, _hash=None):
         """
@@ -626,9 +641,6 @@ class UGrid(object):
         TODO: mask the indices that aren't on the grid properly.
         """
 
-        if not hasattr(self, '_alpha_memo_dict'):
-            self._alpha_memo_dict = OrderedDict()
-
         if _memo:
             if _hash is None:
                 _hash = self._hash_of_pts(points)
@@ -667,23 +679,53 @@ class UGrid(object):
     def interpolate_var_to_points(self,
                                   points,
                                   variable,
+                                  location=None,
+                                  fill_value=0,
                                   indices=None,
-                                  grid=None,
                                   alphas=None,
-                                  mask=None,
                                   slices=None,
-                                  slice_grid=True,
                                   _copy=False,
                                   _memo=True,
                                   _hash=None):
         """
-        interpolates teh passed-in variable to the points in points
+        Interpolates a variable on one of the grids to an array of points.
+        :param points: Nx2 Array of lon/lat coordinates to be interpolated to.
+
+        :param variable: Array-like of values to associate at location on grid (node, center, edge1, edge2).
+        This may be more than a 2 dimensional array, but you must pass 'slices' kwarg with appropriate
+        slice collection to reduce it to 2 dimensions.
+
+        :param location: One of ('node', 'center', 'edge1', 'edge2') 'edge1' is conventionally associated with the
+        'vertical' edges and likewise 'edge2' with the 'horizontal'
+
+        :param fill_value: If masked values are encountered in interpolation, this value takes the place of the masked value
+
+        :param indices: If computed already, array of Nx2 cell indices can be passed in to increase speed. # noqa
+        :param alphas: If computed already, array of alphas can be passed in to increase speed. # noqa
+
+
+        - With a numpy array:
+        sgrid.interpolate_var_to_points(points, sgrid.u[time_idx, depth_idx])
+        - With a raw netCDF Variable:
+        sgrid.interpolate_var_to_points(points, nc.variables['u'], slices=[time_idx, depth_idx])
+
+        If you have pre-computed information, you can pass it in to avoid unnecessary
+        computation and increase performance.
+        - ind = # precomputed indices of points
+        - alphas = # precomputed alphas (useful if interpolating to the same points frequently)
+
+        sgrid.interpolate_var_to_points(points, sgrid.u, indices=ind, alphas=alphas,
+        slices=[time_idx, depth_idx])
 
-        used linear interpolation from the nodes.
         """
         points = np.asarray(points, dtype=np.float64).reshape(-1, 2)
-        location = self.infer_location(variable)
-        # FixMe: should it get location from variable object?
+        # location should be already known by the variable
+        location = variable.location
+        # But if it's not, then it can be infered
+        # (for compatibilty with old code)
+        if location is None:
+            location = self.infer_location(variable)
+            variable.location = location
         if location is None:
             raise ValueError("Data is incompatible with grid nodes or faces")
 
@@ -708,6 +750,8 @@ class UGrid(object):
             return np.sum(vals * pos_alphas, axis=1)
         return None
 
+    interpolate = interpolate_var_to_points
+
     def build_face_face_connectivity(self):
         """
         Builds the face_face_connectivity array: giving the neighbors of each cell.
@@ -871,194 +915,219 @@ class UGrid(object):
             boundary_coordinates[i] = coords.mean(axis=0)
         self.boundary_coordinates = boundary_coordinates
 
-    def save_as_netcdf(self, filepath):
+
+    def save_as_netcdf(self, filename, format='netcdf4'):
+        """
+        save the dataset to a file
+
+        :param filename: full path to file to save to.
+
+        :param format: format to save -- 'netcdf3' or 'netcdf4'
+                       are the only options at this point.
+        """
+        self.save(filename, format='netcdf4')
+
+
+    def save(self, filepath, format='netcdf4', variables={}):
         """
         Save the ugrid object as a netcdf file.
 
         :param filepath: path to file you want o save to.  An existing one
                          will be clobbered if it already exists.
 
+        :param variables: dict of gridded.Variable objects to save to file
+
         Follows the convention established by the netcdf UGRID working group:
 
         http://ugrid-conventions.github.io/ugrid-conventions
 
+        NOTE: Variables are saved here, because different conventions do it
+              differently.
         """
+        format_options = ('netcdf3', 'netcdf4')
+        if format not in format_options:
+            raise ValueError("format: {} not supported. Options are: {}".format(format, format_options))
+
         mesh_name = self.mesh_name
 
-        # dependency only when needed
-        import netCDF4
+        nclocal = get_writable_dataset(filepath)
 
-        # Create a new netcdf file.
-        with netCDF4.Dataset(filepath, mode="w", clobber=True) as nclocal:
+        nclocal.createDimension(mesh_name + "_num_node", len(self.nodes))
+        if self._edges is not None:
+            nclocal.createDimension(
+                mesh_name + "_num_edge", len(self._edges))
+        if self._boundaries is not None:
+            nclocal.createDimension(mesh_name + "_num_boundary",
+                                    len(self._boundaries))
+        if self._faces is not None:
+            nclocal.createDimension(
+                mesh_name + "_num_face", len(self._faces))
+            nclocal.createDimension(mesh_name + "_num_vertices",
+                                    self._faces.shape[1])
+        nclocal.createDimension("two", 2)
 
-            nclocal.createDimension(mesh_name + "_num_node", len(self.nodes))
-            if self._edges is not None:
-                nclocal.createDimension(
-                    mesh_name + "_num_edge", len(self._edges))
-            if self._boundaries is not None:
-                nclocal.createDimension(mesh_name + "_num_boundary",
-                                        len(self._boundaries))
-            if self._faces is not None:
-                nclocal.createDimension(
-                    mesh_name + "_num_face", len(self._faces))
-                nclocal.createDimension(mesh_name + "_num_vertices",
-                                        self._faces.shape[1])
-            nclocal.createDimension("two", 2)
+        # mesh topology
+        mesh = nclocal.createVariable(mesh_name, IND_DT, (),)
+        mesh.cf_role = "mesh_topology"
+        mesh.long_name = "Topology data of 2D unstructured mesh"
+        mesh.topology_dimension = 2
+        mesh.node_coordinates = "{0}_node_lon {0}_node_lat".format(mesh_name)  # noqa
 
-            # mesh topology
-            mesh = nclocal.createVariable(mesh_name, IND_DT, (),)
-            mesh.cf_role = "mesh_topology"
-            mesh.long_name = "Topology data of 2D unstructured mesh"
-            mesh.topology_dimension = 2
-            mesh.node_coordinates = "{0}_node_lon {0}_node_lat".format(mesh_name)  # noqa
-
-            if self._edges is not None:
-                # Attribute required if variables will be defined on edges.
-                mesh.edge_node_connectivity = mesh_name + "_edge_nodes"
-                if self.edge_coordinates is not None:
-                    # Optional attribute (requires edge_node_connectivity).
-                    coord = "{0}_edge_lon {0}_edge_lat".format
-                    mesh.edge_coordinates = coord(mesh_name)
-            if self._faces is not None:
-                mesh.face_node_connectivity = mesh_name + "_face_nodes"
-                if self.face_coordinates is not None:
-                    # Optional attribute.
-                    coord = "{0}_face_lon {0}_face_lat".format
-                    mesh.face_coordinates = coord(mesh_name)
-            if self.face_edge_connectivity is not None:
+        if self._edges is not None:
+            # Attribute required if variables will be defined on edges.
+            mesh.edge_node_connectivity = mesh_name + "_edge_nodes"
+            if self.edge_coordinates is not None:
                 # Optional attribute (requires edge_node_connectivity).
-                mesh.face_edge_connectivity = mesh_name + "_face_edges"
-            if self.face_face_connectivity is not None:
+                coord = "{0}_edge_lon {0}_edge_lat".format
+                mesh.edge_coordinates = coord(mesh_name)
+        if self._faces is not None:
+            mesh.face_node_connectivity = mesh_name + "_face_nodes"
+            if self.face_coordinates is not None:
                 # Optional attribute.
-                mesh.face_face_connectivity = mesh_name + "_face_links"
-            if self._boundaries is not None:
-                mesh.boundary_node_connectivity = mesh_name + "_boundary_nodes"
+                coord = "{0}_face_lon {0}_face_lat".format
+                mesh.face_coordinates = coord(mesh_name)
+        if self.face_edge_connectivity is not None:
+            # Optional attribute (requires edge_node_connectivity).
+            mesh.face_edge_connectivity = mesh_name + "_face_edges"
+        if self.face_face_connectivity is not None:
+            # Optional attribute.
+            mesh.face_face_connectivity = mesh_name + "_face_links"
+        if self._boundaries is not None:
+            mesh.boundary_node_connectivity = mesh_name + "_boundary_nodes"
 
-            # FIXME: This could be re-factored to be more generic, rather than
-            # separate for each type of data see the coordinates example below.
-            if self._faces is not None:
-                nc_create_var = nclocal.createVariable
-                face_nodes = nc_create_var(mesh_name + "_face_nodes", IND_DT,
-                                           (mesh_name + '_num_face',
-                                            mesh_name + '_num_vertices'),)
-                face_nodes[:] = self.faces
+        # FIXME: This could be re-factored to be more generic, rather than
+        # separate for each type of data see the coordinates example below.
+        if self._faces is not None:
+            nc_create_var = nclocal.createVariable
+            face_nodes = nc_create_var(mesh_name + "_face_nodes", IND_DT,
+                                       (mesh_name + '_num_face',
+                                        mesh_name + '_num_vertices'),)
+            face_nodes[:] = self.faces
 
-                face_nodes.cf_role = "face_node_connectivity"
-                face_nodes.long_name = ("Maps every triangular face to "
-                                        "its three corner nodes.")
-                face_nodes.start_index = IND_DT(0)
+            face_nodes.cf_role = "face_node_connectivity"
+            face_nodes.long_name = ("Maps every triangular face to "
+                                    "its three corner nodes.")
+            face_nodes.start_index = IND_DT(0)
 
-            if self._edges is not None:
-                nc_create_var = nclocal.createVariable
-                edge_nodes = nc_create_var(mesh_name + "_edge_nodes", IND_DT,
-                                           (mesh_name + '_num_edge', 'two'),)
-                edge_nodes[:] = self.edges
+        if self._edges is not None:
+            nc_create_var = nclocal.createVariable
+            edge_nodes = nc_create_var(mesh_name + "_edge_nodes", IND_DT,
+                                       (mesh_name + '_num_edge', 'two'),)
+            edge_nodes[:] = self.edges
 
-                edge_nodes.cf_role = "edge_node_connectivity"
-                edge_nodes.long_name = ("Maps every edge to the two "
-                                        "nodes that it connects.")
-                edge_nodes.start_index = IND_DT(0)
+            edge_nodes.cf_role = "edge_node_connectivity"
+            edge_nodes.long_name = ("Maps every edge to the two "
+                                    "nodes that it connects.")
+            edge_nodes.start_index = IND_DT(0)
 
-            if self._boundaries is not None:
-                nc_create_var = nclocal.createVariable
-                boundary_nodes = nc_create_var(mesh_name + "_boundary_nodes",
-                                               IND_DT,
-                                               (mesh_name + '_num_boundary',
-                                                'two'),)
-                boundary_nodes[:] = self.boundaries
+        if self._boundaries is not None:
+            nc_create_var = nclocal.createVariable
+            boundary_nodes = nc_create_var(mesh_name + "_boundary_nodes",
+                                           IND_DT,
+                                           (mesh_name + '_num_boundary',
+                                            'two'),)
+            boundary_nodes[:] = self.boundaries
 
-                boundary_nodes.cf_role = "boundary_node_connectivity"
-                boundary_nodes.long_name = ("Maps every boundary segment to "
-                                            "the two nodes that it connects.")
-                boundary_nodes.start_index = IND_DT(0)
+            boundary_nodes.cf_role = "boundary_node_connectivity"
+            boundary_nodes.long_name = ("Maps every boundary segment to "
+                                        "the two nodes that it connects.")
+            boundary_nodes.start_index = IND_DT(0)
 
-            # Optional "coordinate variables."
-            for location in ['face', 'edge', 'boundary']:
-                loc = "{0}_coordinates".format(location)
-                if getattr(self, loc) is not None:
-                    for axis, ind in [('lat', 1), ('lon', 0)]:
-                        nc_create_var = nclocal.createVariable
-                        name = "{0}_{1}_{2}".format(mesh_name, location, axis)
-                        dimensions = "{0}_num_{1}".format(mesh_name, location)
-                        var = nc_create_var(name, NODE_DT,
-                                            dimensions=(dimensions),)
-                        loc = "{0}_coordinates".format(location)
-                        var[:] = getattr(self, loc)[:, ind]
-                        # Attributes of the variable.
-                        var.standard_name = ("longitude" if axis == 'lon'
-                                             else 'latitude')
-                        var.units = ("degrees_east" if axis == 'lon'
-                                     else 'degrees_north')
-                        name = "Characteristics {0} of 2D mesh {1}".format
-                        var.long_name = name(var.standard_name, location)
+        # Optional "coordinate variables."
+        for location in ['face', 'edge', 'boundary']:
+            loc = "{0}_coordinates".format(location)
+            if getattr(self, loc) is not None:
+                for axis, ind in [('lat', 1), ('lon', 0)]:
+                    nc_create_var = nclocal.createVariable
+                    name = "{0}_{1}_{2}".format(mesh_name, location, axis)
+                    dimensions = "{0}_num_{1}".format(mesh_name, location)
+                    var = nc_create_var(name, NODE_DT,
+                                        dimensions=(dimensions),)
+                    loc = "{0}_coordinates".format(location)
+                    var[:] = getattr(self, loc)[:, ind]
+                    # Attributes of the variable.
+                    var.standard_name = ("longitude" if axis == 'lon'
+                                         else 'latitude')
+                    var.units = ("degrees_east" if axis == 'lon'
+                                 else 'degrees_north')
+                    name = "Characteristics {0} of 2D mesh {1}".format
+                    var.long_name = name(var.standard_name, location)
 
-            # The node data.
-            node_lon = nclocal.createVariable(mesh_name + '_node_lon',
-                                              self._nodes.dtype,
-                                              (mesh_name + '_num_node',),
-                                              chunksizes=(len(self.nodes),),
+        # The node data.
+        node_lon = nclocal.createVariable(mesh_name + '_node_lon',
+                                          self._nodes.dtype,
+                                          (mesh_name + '_num_node',),
+                                          chunksizes=(len(self.nodes),),
+                                          # zlib=False,
+                                          # complevel=0,
+                                          )
+        node_lon[:] = self.nodes[:, 0]
+        node_lon.standard_name = "longitude"
+        node_lon.long_name = "Longitude of 2D mesh nodes."
+        node_lon.units = "degrees_east"
+
+        node_lat = nclocal.createVariable(mesh_name + '_node_lat',
+                                          self._nodes.dtype,
+                                          (mesh_name + '_num_node',),
+                                          chunksizes=(len(self.nodes),),
+                                          # zlib=False,
+                                          # complevel=0,
+                                          )
+        node_lat[:] = self.nodes[:, 1]
+        node_lat.standard_name = "latitude"
+        node_lat.long_name = "Latitude of 2D mesh nodes."
+        node_lat.units = "degrees_north"
+
+        self._save_variables(nclocal, variables)
+
+        nclocal.sync()
+        return nclocal
+
+    def _save_variables(self, nclocal, variables):
+        """
+        Save the Variables
+        """
+        mesh_name = self.mesh_name
+        for name, var in variables.items():
+            if var.location == 'node':
+                shape = (mesh_name + '_num_node',)
+                coordinates = "{0}_node_lon {0}_node_lat".format(mesh_name)
+                chunksizes = (len(self.nodes),)
+            elif var.location == 'face':
+                shape = (mesh_name + '_num_face',)
+                coord = "{0}_face_lon {0}_face_lat".format
+                coordinates = (coord(mesh_name) if self.face_coordinates
+                               is not None else None)
+                chunksizes = (len(self.faces),)
+            elif var.location == 'edge':
+                shape = (mesh_name + '_num_edge',)
+                coord = "{0}_edge_lon {0}_edge_lat".format
+                coordinates = (coord(mesh_name) if self.edge_coordinates
+                               is not None else None)
+                chunksizes = (len(self.edges),)
+            elif var.location == 'boundary':
+                shape = (mesh_name + '_num_boundary',)
+                coord = "{0}_boundary_lon {0}_boundary_lat".format
+                bcoord = self.boundary_coordinates
+                coordinates = (coord(mesh_name) if bcoord
+                               is not None else None)
+                chunksizes = (len(self.boundaries),)
+            else:
+                raise ValueError("I dont' know how to save a variable located on: {}".format(var.location))
+            data_var = nclocal.createVariable(var.name,
+                                              var.data.dtype,
+                                              shape,
+                                              chunksizes=chunksizes,
                                               # zlib=False,
                                               # complevel=0,
                                               )
-            node_lon[:] = self.nodes[:, 0]
-            node_lon.standard_name = "longitude"
-            node_lon.long_name = "Longitude of 2D mesh nodes."
-            node_lon.units = "degrees_east"
+            data_var[:] = var.data
+            # Add the standard attributes:
+            data_var.location = var.location
+            data_var.mesh = mesh_name
+            if coordinates is not None:
+                data_var.coordinates = coordinates
+            # Add the extra attributes.
+            for att_name, att_value in var.attributes.items():
+                setattr(data_var, att_name, att_value)
 
-            node_lat = nclocal.createVariable(mesh_name + '_node_lat',
-                                              self._nodes.dtype,
-                                              (mesh_name + '_num_node',),
-                                              chunksizes=(len(self.nodes),),
-                                              # zlib=False,
-                                              # complevel=0,
-                                              )
-            node_lat[:] = self.nodes[:, 1]
-            node_lat.standard_name = "latitude"
-            node_lat.long_name = "Latitude of 2D mesh nodes."
-            node_lat.units = "degrees_north"
-
-            # This should be done from the gridded.Dataset now.
-            #  The variables should know how to write themselves
-            # # Write the associated data.
-            # for dataset in self.data.values():
-            #     if dataset.location == 'node':
-            #         shape = (mesh_name + '_num_node',)
-            #         coordinates = "{0}_node_lon {0}_node_lat".format(mesh_name)
-            #         chunksizes = (len(self.nodes),)
-            #     elif dataset.location == 'face':
-            #         shape = (mesh_name + '_num_face',)
-            #         coord = "{0}_face_lon {0}_face_lat".format
-            #         coordinates = (coord(mesh_name) if self.face_coordinates
-            #                        is not None else None)
-            #         chunksizes = (len(self.faces),)
-            #     elif dataset.location == 'edge':
-            #         shape = (mesh_name + '_num_edge',)
-            #         coord = "{0}_edge_lon {0}_edge_lat".format
-            #         coordinates = (coord(mesh_name) if self.edge_coordinates
-            #                        is not None else None)
-            #         chunksizes = (len(self.edges),)
-            #     elif dataset.location == 'boundary':
-            #         shape = (mesh_name + '_num_boundary',)
-            #         coord = "{0}_boundary_lon {0}_boundary_lat".format
-            #         bcoord = self.boundary_coordinates
-            #         coordinates = (coord(mesh_name) if bcoord
-            #                        is not None else None)
-            #         chunksizes = (len(self.boundaries),)
-            #     data_var = nclocal.createVariable(dataset.name,
-            #                                       dataset.data.dtype,
-            #                                       shape,
-            #                                       chunksizes=chunksizes,
-            #                                       # zlib=False,
-            #                                       # complevel=0,
-            #                                       )
-            #     data_var[:] = dataset.data
-            #     # Add the standard attributes:
-            #     data_var.location = dataset.location
-            #     data_var.mesh = mesh_name
-            #     if coordinates is not None:
-            #         data_var.coordinates = coordinates
-            #     # Add the extra attributes.
-            #     for att_name, att_value in dataset.attributes.items():
-            #         setattr(data_var, att_name, att_value)
-
-            nclocal.sync()
