@@ -9,8 +9,11 @@ import os
 import numpy as np
 import netCDF4 as nc4
 
-from gridded.utilities import get_dataset, _reorganize_spatial_data,\
-    _align_results_to_spatial_data
+from gridded.utilities import (get_dataset,
+                               _reorganize_spatial_data,
+                               _align_results_to_spatial_data,
+                               asarraylike)
+from gridded import VALID_LOCATIONS
 from gridded.grids import Grid, Grid_U, Grid_S, Grid_R
 from gridded.depth import Depth
 from gridded.time import Time
@@ -54,6 +57,7 @@ class Variable(object):
                  dataset=None,
                  varname=None,
                  fill_value=0,
+                 location=None,
                  attributes=None,
                  **kwargs):
         '''
@@ -85,7 +89,11 @@ class Variable(object):
 
         :param fill_value: the fill value used for undefined data
 
-        :param attributes: attributes associated with the VAriable
+        :param location: location on the grid -- possible values
+                         depend on the grid type
+        :type location: str
+
+        :param attributes: attributes associated with the Variable
                            (analogous to netcdf variable attributes)
         :type attributes: dict of key:value pairs
         '''
@@ -102,6 +110,7 @@ class Variable(object):
 
         self.name = name
         self.units = units
+        self.location = location
         self.data = data
         self.time = time if time is not None else self._default_component_types['time'].constant_time()
         self.data_file = data_file
@@ -138,25 +147,41 @@ class Variable(object):
                     dataset=None,
                     data_file=None,
                     grid_file=None,
-                    load_all=False,
+                    location=None,
+                    load_all=False,  # Do we need this? I think not --- maybe a method to fully load later if wanted.
                     fill_value=0,
                     **kwargs
                     ):
         '''
         Allows one-function creation of a Variable from a file.
 
-        :param filename: Default data source. Parameters below take precedence
-        :param varname: Name of the variable in the data source file
-        :param grid_topology: Description of the relationship between grid attributes and variable names.
-        :param name: Name of property
-        :param units: Units
-        :param time: Time axis of the data
-        :param data: Underlying data source
+        :param filename: Default data source. Has lowest priority; 
+        if dataset, grid_file, or data_file are provided, this function 
+        uses them first
+
+        :param varname: Explicit name of the data in the data source file.
+        Equivalent to the key used to look the item up directly eg 'ds["lon_u"]'
+        for a netCDF4 Dataset.
+
+        :param grid_topology: Description of the relationship between grid attributes
+        and variable names.
+        :param name: Name of this object
+        :param units: string such as 'm/s'
+
+        :param time: Time axis of the data. May be a constructed gridded.Time
+        object, or collection of datetime.datetime objects
+
+        :param data: Underlying data object. May be any array-like,
+        including netCDF4 Variable, etc
+
         :param grid: Grid that the data corresponds with
-        :param depth: Depth axis object
-        :param dataset: Instance of open Dataset
-        :param data_file: Name of data source file
-        :param grid_file: Name of grid source file
+        :param location: The feature where the data aligns with the grid.
+
+        :param depth: Depth axis object from gridded.depth
+
+        :param dataset: Instance of open netCDF4.Dataset
+        :param data_file: Name of data source file, if data and grid files are separate
+        :param grid_file: Name of grid source file, if data and grid files are separate
         :type filename: string
         :type varname: string
         :type grid_topology: {string : string, ...}
@@ -165,6 +190,7 @@ class Variable(object):
         :type time: [] of datetime.datetime, netCDF4 Variable, or Time object
         :type data: netCDF4.Variable or numpy.array
         :type grid: pysgrid or pyugrid
+        :type location: string
         :type depth: Depth, S_Depth or L_Depth
         :type dataset: netCDF4.Dataset
         :type data_file: string
@@ -228,6 +254,9 @@ class Variable(object):
                 depth = Depth.from_netCDF(grid_file,
                                           dataset=dg,
                                           )
+        if location is None:
+            if hasattr(data, 'location'):
+                location = data.location
 #             if len(data.shape) == 4 or (len(data.shape) == 3 and time is None):
 #                 from gnome.environment.environment_objects import S_Depth
 #                 depth = S_Depth.from_netCDF(grid=grid,
@@ -246,6 +275,7 @@ class Variable(object):
                    grid_file=grid_file,
                    data_file=data_file,
                    fill_value=fill_value,
+                   location=location,
                    varname=varname,
                    **kwargs)
 
@@ -260,6 +290,21 @@ class Variable(object):
                 'data="{0.data}", '
                 ')').format(self)
 
+    @property
+    def location(self):
+        if self._location is None and self.data is not None and hasattr(self.data, 'location'):
+            return self.data.location
+        else:
+            return self._location
+
+    @location.setter
+    def location(self, location):
+        # Fixme: perhaps we need Variable subclasses,
+        #        to distingish between variable types.
+        if location not in VALID_LOCATIONS:
+            raise ValueError("Invalid location: {}, must be one of: {}".format(location, VALID_LOCATIONS))
+        self._location = location
+
     @property
     def info(self):
         """
@@ -306,12 +351,22 @@ class Variable(object):
 
     @data.setter
     def data(self, d):
+        d = asarraylike(d)
+        # Fixme: maybe all this checking should be done when it gets added to the Dataset??
         if self.time is not None and len(d) != len(self.time):
             raise ValueError("Data/time interval mismatch")
-        if self.grid is not None and self.grid.infer_location(d) is None:
-            raise ValueError("Data/grid shape mismatch. Data shape is {0}, Grid shape is {1}".format(d.shape, self.grid.node_lon.shape))
+        ## fixme: we should check Depth, too.
+        # if self.grid is not None and self.grid.infer_location(d) is None:
+        #     raise ValueError("Data/grid shape mismatch. Data shape is {0}, Grid shape is {1}".format(d.shape, self.grid.node_lon.shape))
+        if self.grid is not None:  # if there is not a grid, we can't check this
+            if self.location is None:  # not set, let's try to figure it out
+                self.location = self.grid.infer_location(d)
+            if self.location is None:
+                raise ValueError("Data/grid shape mismatch: Data shape is {0}, "
+                                 "Grid shape is {1}".format(d.shape, self.grid.node_lon.shape))
         self._data = d
 
+
     @property
     def units(self):
         '''
@@ -504,6 +559,8 @@ class Variable(object):
             self._memoize_result(pts, time, value, self._result_memo, _hash=_hash)
         return value
 
+    interpolate = at #common request
+
     def _xy_interp(self, points, time, extrapolate, slices=(), **kwargs):
         '''
         Uses the py(s/u)grid interpolation to determine the values at the points, and returns it
@@ -521,6 +578,7 @@ class Variable(object):
 
         value = self.grid.interpolate_var_to_points(points[:, 0:2],
                                                     self.data,
+                                                    location=self.location,
                                                     _hash=self._get_hash(points[:, 0:2],
                                                                          time),
                                                     slices=slices, _memo=True)
@@ -676,6 +734,7 @@ class VectorVariable(object):
                  varnames=None,
                  **kwargs):
 
+        super(VectorVariable, self).__init__()
         self.name = self._units = self._time = self._variables = None
 
         self.name = name
@@ -774,8 +833,8 @@ class VectorVariable(object):
 
         if grid is None:
             grid = Grid.from_netCDF(grid_file,
-                                      dataset=dg,
-                                      grid_topology=grid_topology)
+                                    dataset=dg,
+                                    grid_topology=grid_topology)
         if varnames is None:
             varnames = cls._gen_varnames(data_file,
                                          dataset=ds)
@@ -813,17 +872,19 @@ class VectorVariable(object):
             variables = []
             for vn in varnames:
                 if vn is not None:
+                    # Fixme: We're calling from_netCDF from itself ?!?!?
                     variables.append(Variable.from_netCDF(filename=filename,
                                                           varname=vn,
                                                           grid_topology=grid_topology,
                                                           units=units,
                                                           time=time,
                                                           grid=grid,
-                                                           depth=depth,
+                                                          depth=depth,
                                                           data_file=data_file,
                                                           grid_file=grid_file,
                                                           dataset=ds,
                                                           load_all=load_all,
+                                                          location=None,
                                                           **kwargs))
         if units is None:
             units = [v.units for v in variables]
@@ -842,6 +903,7 @@ class VectorVariable(object):
                    grid_file=grid_file,
                    dataset=ds,
                    load_all=load_all,
+                   location=None,
                    **kwargs)
 
     @classmethod
@@ -900,6 +962,12 @@ class VectorVariable(object):
                 'grid="{0.grid}", '
                 ')').format(self)
 
+    @property
+    def location(self):
+        return [v.location for v in self.variables]
+    
+    locations = location
+
     @property
     def is_data_on_nodes(self):
         return self.grid.infer_location(self.variables[0].data) == 'node'
@@ -1063,3 +1131,21 @@ class VectorVariable(object):
                 return func(*args, **kws)
             return wrapper
         return getvars
+
+    def save(self, filepath, format='netcdf4'):
+        """
+        Save the variable object to a netcdf file.
+
+        :param filepath: path to file you want o save to. or a writable
+                         netCDF4 Dataset An existing one
+                         If a path, an existing file will be clobbered.
+
+        Follows the convention established by the netcdf UGRID working group:
+
+        http://ugrid-conventions.github.io/ugrid-conventions
+
+        """
+        format_options = ('netcdf3', 'netcdf4')
+        if format not in format_options:
+            raise ValueError("format: {} not supported. Options are: {}".format(format, format_options))
+
